from unittest import TestCase
from datetime import date

from cobalt import Act, datestring
from cobalt.schemas import assert_validates


class StructuredDocumentTestCase(TestCase):
    # using Act to test StructuredDocument functionality
    # because we need a concrete document (with a default empty_document) to work with
    maxDiff = None

    def test_frbr_uri(self):
        a = Act()
        a.expression_date = '2012-01-01'
        a.frbr_uri = '/zm/act/2007/01'
        today = datestring(date.today())

        self.assertEqual(a.frbr_uri.work_uri(), '/zm/act/2007/01')

        self.assertEqual(a.meta.identification.FRBRWork.FRBRthis.get('value'), '/zm/act/2007/01/!main')
        self.assertEqual(a.meta.identification.FRBRWork.FRBRuri.get('value'), '/zm/act/2007/01')
        self.assertEqual(a.meta.identification.FRBRWork.FRBRcountry.get('value'), 'zm')
        self.assertEqual(a.meta.identification.FRBRWork.FRBRnumber.get('value'), '01')
        self.assertEqual(a.meta.identification.FRBRWork.FRBRdate.get('date'), '2007')

        self.assertEqual(a.meta.identification.FRBRExpression.FRBRthis.get('value'), '/zm/act/2007/01/eng@2012-01-01/!main')
        self.assertEqual(a.meta.identification.FRBRExpression.FRBRuri.get('value'), '/zm/act/2007/01/eng@2012-01-01')
        self.assertEqual(a.meta.identification.FRBRExpression.FRBRdate.get('date'), '2012-01-01')

        self.assertEqual(a.meta.identification.FRBRManifestation.FRBRthis.get('value'),
                     '/zm/act/2007/01/eng@2012-01-01/!main')
        self.assertEqual(a.meta.identification.FRBRManifestation.FRBRuri.get('value'), '/zm/act/2007/01/eng@2012-01-01')
        self.assertEqual(a.meta.identification.FRBRManifestation.FRBRdate.get('date'), today)

        assert_validates(a)

    def test_frbr_country(self):
        a = Act()
        a.expression_date = '2012-01-01'
        a.frbr_uri = '/zm/act/2007/01'

        self.assertEqual(a.meta.identification.FRBRWork.FRBRcountry.get('value'), 'zm')

        a.frbr_uri = '/zm-abc123/act/2007/01'
        self.assertEqual(a.meta.identification.FRBRWork.FRBRcountry.get('value'), 'zm-abc123')

    def test_title(self):
        a = Act()
        a.title = "a title"
        self.assertEqual(a.title, "a title")

    def test_subtype(self):
        a = Act()
        a.frbr_uri = '/akn/za/act/by-law/2009/1'
        self.assertEqual(a.meta.identification.FRBRWork.FRBRsubtype.get('value'), 'by-law')
        assert_validates(a)

        # clear it
        a.frbr_uri = '/akn/za/act/2009/1'
        with self.assertRaises(AttributeError):
            a.meta.identification.FRBRWork.FRBRsubtype

        assert_validates(a)

    def test_work_date(self):
        """ Work date must be the FRBR date.
        """
        a = Act()
        a.frbr_uri = '/akn/za/act/2012-01-02/5'
        self.assertEqual(datestring(a.work_date), '2012-01-02')
        self.assertIsInstance(a.work_date, date)

        a.frbr_uri = '/akn/za/act/2009/5'
        self.assertEqual(datestring(a.work_date), '2009-01-01')
        self.assertIsInstance(a.work_date, date)

    def test_expression_date(self):
        a = Act()
        a.expression_date = '2012-01-02'
        self.assertEqual(datestring(a.expression_date), '2012-01-02')
        self.assertIsInstance(a.expression_date, date)

    def test_manifestation_date(self):
        a = Act()
        a.manifestation_date = '2012-01-02'
        self.assertEqual(datestring(a.manifestation_date), '2012-01-02')
        self.assertIsInstance(a.manifestation_date, date)

    def test_language(self):
        a = Act()
        a.language = 'fre'
        self.assertEqual(a.language, 'fre')

    def test_namespaces(self):
        # default is now AKN3
        a = Act()
        self.assertEqual(a.namespace, 'http://docs.oasis-open.org/legaldocml/ns/akn/3.0')

        # prefer AKN3 when both 2 and 3 are listed as namespaces
        a = Act(xml="""<?xml version="1.0"?>
<foo:akomaNtoso xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:foo="http://www.akomantoso.org/2.0" xmlns:bar="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xsi:schemaLocation="http://www.akomantoso.org/2.0 akomantoso20.xsd">
  <foo:act>
    <meta/>
    <foo:body/>
  </foo:act>
  <bar:act>
    <meta/>
    <bar:body/>
  </bar:act>
</foo:akomaNtoso>""")
        self.assertEqual(a.namespace, 'http://docs.oasis-open.org/legaldocml/ns/akn/3.0')

        # prefer AKN2 when 2 and something else are listed as namespaces
        a = Act(xml="""<?xml version="1.0"?>
<foo:akomaNtoso xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:foo="http://www.akomantoso.org/2.0" xmlns:bar="http://docs.oasis-open.org/legaldocml/ns/akn/5.0" xsi:schemaLocation="http://www.akomantoso.org/2.0 akomantoso20.xsd">
  <foo:act>
    <meta/>
    <foo:body/>
  </foo:act>
</foo:akomaNtoso>""")
        self.assertEqual(a.namespace, 'http://www.akomantoso.org/2.0')

        # throw error if neither of AKN2 and AKN3 are listed as namespaces
        with self.assertRaises(ValueError) as raised:
            Act(xml="""<?xml version="1.0"?>
                <foo:akomaNtoso xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:foo="http://www.akomantoso.org/4.0" xmlns:bar="http://docs.oasis-open.org/legaldocml/ns/akn/5.0" xsi:schemaLocation="http://www.akomantoso.org/2.0 akomantoso20.xsd">
  <foo:act>
    <meta/>
    <body/>
  </foo:act>
</foo:akomaNtoso>""")
        self.assertIn(
            "Expected to find one of the following Akoma Ntoso XML namespaces: http://docs.oasis-open.org/legaldocml/ns/akn/3.0, http://www.akomantoso.org/2.0. Only these namespaces were found: http://www.w3.org/2001/XMLSchema-instance, http://www.akomantoso.org/4.0, http://docs.oasis-open.org/legaldocml/ns/akn/5.0",
            raised.exception.args)

        # ignore content that isn't in the chosen namespace
        a = Act(xml="""<?xml version="1.0"?>
<bar:akomaNtoso xmlns:foo="http://www.akomantoso.org/2.0" xmlns:bar="http://docs.oasis-open.org/legaldocml/ns/akn/3.0">
  <bar:act>
    <bar:meta/>
    <bar:body>
      <bar:section id="section-1">
        <bar:content>
          <p>This content SHOULD be ignored as it's not using the AKN3 namespace.</p>
          <foo:p>This content should ALSO be ignored as it's using the AKN2 namespace.</foo:p>
          <bar:p>This content should NOT be ignored as it's using the AKN3 namespace.</bar:p>
        </bar:content>
      </bar:section>
    </bar:body>
  </bar:act>
</bar:akomaNtoso>""")
        self.assertEqual(a.namespace, 'http://docs.oasis-open.org/legaldocml/ns/akn/3.0')
        self.assertEqual("This content should NOT be ignored as it's using the AKN3 namespace.", a.body.section.content.p)
        self.assertNotIn("This content SHOULD be ignored as it's not using the AKN3 namespace.", a.body.section.content.p)
        self.assertNotIn("This content should ALSO be ignored as it's using the AKN2 namespace.", a.body.section.content.p)

    def test_parser(self):
        a = Act()
        # no errors raised by parsing default Act
        a.parse(a.to_xml(), a.document_type)

        # error if root isn't `akomaNtoso`
        with self.assertRaises(ValueError) as raised:
            a.parse("""<?xml version="1.0"?>
<myBlog xmlns="http://www.akomantoso.org/2.0">
  <p>Whaddup, fam!</p>
</myBlog>""", a.document_type)
        self.assertIn("XML root element must be akomaNtoso, but got myBlog instead", raised.exception.args)

        # error if root as no children
        with self.assertRaises(ValueError) as raised:
            a.parse("""<?xml version="1.0"?>
<akomaNtoso xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xsi:schemaLocation="http://docs.oasis-open.org/legaldocml/akn-core/v1.0/os/part2-specs/schemas/akomantoso30.xsd">
</akomaNtoso>""", a.document_type)
        self.assertIn("Expected act as a child of root element", raised.exception.args)

        # error if `act` isn't first child
        with self.assertRaises(ValueError) as raised:
            a.parse("""<?xml version="1.0"?>
<akomaNtoso xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xsi:schemaLocation="http://docs.oasis-open.org/legaldocml/akn-core/v1.0/os/part2-specs/schemas/akomantoso30.xsd">
  <somethingElse>
  </somethingElse>
</akomaNtoso>""", a.document_type)
        self.assertIn("Expected act as a child of root element", raised.exception.args)

        # allow comments at the top level
        a.parse("""<akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0">
  <!-- a comment -->
  <act>
    <meta/>
    <body/>
  </act>
</akomaNtoso>""", a.document_type)

    def test_unicode(self):
        # string, no encoding
        a = Act("""<akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0">
          <act>
            <meta/>
            <body>😀</body>
          </act>
        </akomaNtoso>""")
        self.assertEqual(a.root.xpath("//a:body", namespaces={'a': a.namespace})[0].text, "😀")

        # bytes, no encoding
        a = Act("""<akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0">
          <act>
            <meta/>
            <body>😀</body>
          </act>
        </akomaNtoso>""".encode('utf-8'))
        self.assertEqual(a.root.xpath("//a:body", namespaces={'a': a.namespace})[0].text, "😀")

        # with encoding attribute, bytes
        a = Act("""<?xml version="1.0" encoding="utf-8"?><akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0">
          <act>
            <meta/>
            <body>😀</body>
          </act>
        </akomaNtoso>""".encode('utf-8'))
        self.assertEqual(a.root.xpath("//a:body", namespaces={'a': a.namespace})[0].text, "😀")

        # with encoding string
        Act("""<?xml version="1.0" encoding="utf-8"?><akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0">
          <act>
            <meta/>
            <body>😀</body>
          </act>
        </akomaNtoso>""")

    def test_add_number(self):
        """ When adding an FRBRnumber element to a document that doesn't already have one, it
        must come after subtype.
        """
        a = Act(xml="""
<akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xsi:schemaLocation="http://docs.oasis-open.org/legaldocml/akn-core/v1.0/os/part2-specs/schemas/akomantoso30.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <!-- a comment -->
  <act contains="singleVersion" name="act">
    <meta>
      <identification source="#slaw">
        <FRBRWork>
          <FRBRthis value="/na/act/p/1977/25/!main"/>
          <FRBRuri value="/na/act/p/1977/25"/>
          <FRBRalias value="Livestock Improvement Act, 1977" name="title"/>
          <FRBRdate date="1977-03-23" name="Generation"/>
          <FRBRauthor href="#council"/>
          <FRBRcountry value="na"/>
          <FRBRsubtype value="p"/>
        </FRBRWork>
        <FRBRExpression>
          <FRBRthis value="/na/act/p/1977/25/eng@1993-12-02/!main"/>
          <FRBRuri value="/na/act/p/1977/25/eng@1993-12-02"/>
          <FRBRdate date="1993-12-02" name="Generation"/>
          <FRBRauthor href="#council"/>
          <FRBRlanguage language="eng"/>
        </FRBRExpression>
        <FRBRManifestation>
          <FRBRthis value="/na/act/p/1977/25/eng@1993-12-02/!main"/>
          <FRBRuri value="/na/act/p/1977/25/eng@1993-12-02"/>
          <FRBRdate date="2020-03-25" name="Generation"/>
          <FRBRauthor href="#slaw"/>
        </FRBRManifestation>
      </identification>
      <publication number="5462" name="South African Government Gazette" showAs="South African Government Gazette" date="1977-03-23"/>
    </meta>
    <body>
      <section eId="section_1">
        <content>
          <p></p>
        </content>
      </section>
    </body>
  </act>
</akomaNtoso>
""")
        a.frbr_uri = '/na/act/p/1997/25/'

        assert_validates(a)



# -*- coding: utf-8 -*-
#
# Cobalt documentation build configuration file.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('..'))

# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx_rtd_theme',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# General information about the project.
project = u'Cobalt'
copyright = u'2015-2020, AfricanLII'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
with open("../VERSION") as f:
    release = f.read().strip()

# The short X.Y version.
version = '.'.join(release.split('.')[:2])

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'



from unittest import TestCase

from cobalt import Debate


class DebateTestCase(TestCase):
    maxDiff = None

    def test_empty_debate(self):
        d = Debate()
        self.assertEqual(d.title, "Untitled")
        self.assertIsNotNone(d.meta)
        self.assertIsNotNone(d.debateBody)

    def test_empty_body(self):
        d = Debate()
        self.assertNotEqual(d.debateBody.text, '')

    def test_main(self):
        d = Debate()
        self.assertEqual(d.main, d.debate)



from unittest import TestCase

from lxml import etree

from cobalt import Act
from cobalt.schemas import assert_validates


class AttachmentsTestCase(TestCase):
    maxDiff = None

    def tostring(self, xml):
        return etree.tostring(xml, encoding='unicode').strip()

    def setUp(self):
        self.a = Act(xml="""<?xml version="1.0" encoding="UTF-8"?>
<akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xsi:schemaLocation="http://docs.oasis-open.org/legaldocml/akn-core/v1.0/os/part2-specs/schemas/akomantoso30.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <act contains="singleVersion" name="act">
    <meta>
      <identification source="#cobalt">
        <FRBRWork>
          <FRBRthis value="/na/act/1977/25/!main"/>
          <FRBRuri value="/na/act/1977/25"/>
          <FRBRalias value="Livestock Improvement Act, 1977" name="title"/>
          <FRBRdate date="1977" name="Generation"/>
          <FRBRauthor href=""/>
          <FRBRcountry value="na"/>
          <FRBRnumber value="25"/>
        </FRBRWork>
        <FRBRExpression>
          <FRBRthis value="/na/act/1977/25/eng@1993-12-02/!main"/>
          <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
          <FRBRdate date="1993-12-02" name="Generation"/>
          <FRBRauthor href=""/>
          <FRBRlanguage language="eng"/>
        </FRBRExpression>
        <FRBRManifestation>
          <FRBRthis value="/na/act/1977/25/eng@1993-12-02/!main"/>
          <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
          <FRBRdate date="2020-03-25" name="Generation"/>
          <FRBRauthor href=""/>
        </FRBRManifestation>
      </identification>
      <publication number="5462" name="South African Government Gazette" showAs="South African Government Gazette" date="1977-03-23"/>
    </meta>
    <body>
      <section eId="section_1">
        <content>
          <p></p>
        </content>
      </section>
    </body>
    <attachments>
      <attachment eId="att_1">
        <heading>Schedule</heading>
        <doc name="schedule">
          <meta>
            <identification source="#cobalt">
              <FRBRWork>
                <FRBRthis value="/na/act/1977/25/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25"/>
                <FRBRalias value="Schedule" name="title"/>
                <FRBRdate date="1977" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRcountry value="na"/>
                <FRBRnumber value="25"/>
              </FRBRWork>
              <FRBRExpression>
                <FRBRthis value="/na/act/1977/25/eng@1993-12-02/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                <FRBRdate date="1993-12-02" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRlanguage language="eng"/>
              </FRBRExpression>
              <FRBRManifestation>
                <FRBRthis value="/na/act/1977/25/eng@1993-12-02/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                <FRBRdate date="2020-03-25" name="Generation"/>
                <FRBRauthor href=""/>
              </FRBRManifestation>
            </identification>
          </meta>
          <mainBody>
            <paragraph eId="paragraph_1">
              <content>
                <p>This is the content of the Schedule!</p>
              </content>
            </paragraph>
          </mainBody>
        </doc>
      </attachment>
    </attachments>
    <components>
      <component eId="comp_1">
        <heading>Schedule</heading>
        <doc name="schedule">
          <meta>
            <identification source="#cobalt">
              <FRBRWork>
                <FRBRthis value="/na/act/1977/25/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25"/>
                <FRBRalias value="Schedule" name="title" />
                <FRBRdate date="1980-01-01" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRcountry value="na"/>
                <FRBRnumber value="25"/>
              </FRBRWork>
              <FRBRExpression>
                <FRBRthis value="/na/act/1977/25/eng@1993-12-02/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                <FRBRdate date="1980-01-01" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRlanguage language="eng"/>
              </FRBRExpression>
              <FRBRManifestation>
                <FRBRthis value="/na/act/1977/25/eng@1993-12-02/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                <FRBRdate date="2020-03-25" name="Generation"/>
                <FRBRauthor href=""/>
              </FRBRManifestation>
            </identification>
          </meta>
          <mainBody>
            <paragraph eId="paragraph_1">
              <content>
                <p>This is the content of the Schedule!</p>
              </content>
            </paragraph>
          </mainBody>
        </doc>
      </component>
    </components>
  </act>
</akomaNtoso>
        """)

    def test_component_basics(self):
        components = self.a.components()
        self.assertEqual(['main', 'schedule-A', 'schedule-XXX'], sorted(components.keys()))
        self.assertEqual('This is the content of the Schedule!',
                         components['schedule-XXX'].doc.mainBody.paragraph.content.p)
        assert_validates(self.a)

        # this won't change any component names, because the existing names are used by default
        frbr_uri = self.a.frbr_uri
        frbr_uri.work_component = 'blah'
        self.a.frbr_uri = frbr_uri
        self.assertEqual(['main', 'schedule-A', 'schedule-XXX'], sorted(self.a.components().keys()))

    def test_set_missing_component_with_default(self):
        a = Act(xml="""<?xml version="1.0" encoding="UTF-8"?>
        <akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xsi:schemaLocation="http://docs.oasis-open.org/legaldocml/akn-core/v1.0/os/part2-specs/schemas/akomantoso30.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
          <act contains="singleVersion" name="act">
            <meta>
              <identification source="#cobalt">
                <FRBRWork>
                  <FRBRthis value="/na/act/1977/25"/>
                  <FRBRuri value="/na/act/1977/25"/>
                  <FRBRalias value="Livestock Improvement Act, 1977" name="title"/>
                  <FRBRdate date="1977-03-23" name="Generation"/>
                  <FRBRauthor href=""/>
                  <FRBRcountry value="na"/>
                </FRBRWork>
                <FRBRExpression>
                  <FRBRthis value="/na/act/1977/25/eng@1993-12-02"/>
                  <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                  <FRBRdate date="1993-12-02" name="Generation"/>
                  <FRBRauthor href=""/>
                  <FRBRlanguage language="eng"/>
                </FRBRExpression>
                <FRBRManifestation>
                  <FRBRthis value="/na/act/1977/25/eng@1993-12-02"/>
                  <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                  <FRBRdate date="2020-03-25" name="Generation"/>
                  <FRBRauthor href=""/>
                </FRBRManifestation>
              </identification>
              <publication number="5462" name="South African Government Gazette" showAs="South African Government Gazette" date="1977-03-23"/>
            </meta>
            <body>
              <section eId="section_1">
                <content>
                  <p></p>
                </content>
              </section>
            </body>
          </act>
        </akomaNtoso>
                """)
        self.assertEqual([None], sorted(a.components().keys()))
        assert_validates(a)

        # change the main component name implicitly
        # should get 'main' as the default
        frbr_uri = a.frbr_uri
        frbr_uri.work_component = None
        a.frbr_uri = frbr_uri
        self.assertEqual(['main'], sorted(a.components().keys()))

    def test_set_missing_component_explicitly(self):
        a = Act(xml="""<?xml version="1.0" encoding="UTF-8"?>
        <akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xsi:schemaLocation="http://docs.oasis-open.org/legaldocml/akn-core/v1.0/os/part2-specs/schemas/akomantoso30.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
          <act contains="singleVersion" name="act">
            <meta>
              <identification source="#cobalt">
                <FRBRWork>
                  <FRBRthis value="/na/act/1977/25"/>
                  <FRBRuri value="/na/act/1977/25"/>
                  <FRBRalias value="Livestock Improvement Act, 1977" name="title"/>
                  <FRBRdate date="1977-03-23" name="Generation"/>
                  <FRBRauthor href=""/>
                  <FRBRcountry value="na"/>
                </FRBRWork>
                <FRBRExpression>
                  <FRBRthis value="/na/act/1977/25/eng@1993-12-02"/>
                  <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                  <FRBRdate date="1993-12-02" name="Generation"/>
                  <FRBRauthor href=""/>
                  <FRBRlanguage language="eng"/>
                </FRBRExpression>
                <FRBRManifestation>
                  <FRBRthis value="/na/act/1977/25/eng@1993-12-02"/>
                  <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                  <FRBRdate date="2020-03-25" name="Generation"/>
                  <FRBRauthor href=""/>
                </FRBRManifestation>
              </identification>
              <publication number="5462" name="South African Government Gazette" showAs="South African Government Gazette" date="1977-03-23"/>
            </meta>
            <body>
              <section eId="section_1">
                <content>
                  <p></p>
                </content>
              </section>
            </body>
          </act>
        </akomaNtoso>
                """)
        self.assertEqual([None], sorted(a.components().keys()))
        assert_validates(a)

        # change the main component name
        frbr_uri = a.frbr_uri
        frbr_uri.work_component = 'blah'
        a.frbr_uri = frbr_uri
        self.assertEqual(['blah'], sorted(a.components().keys()))

    def test_component_expression_date(self):
        self.a.expression_date = '2021-01-01'

        components = list(self.a.components().values())

        self.assertEqual('''<meta xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
            <identification source="#cobalt">
              <FRBRWork>
                <FRBRthis value="/na/act/1977/25/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25"/>
                <FRBRalias value="Schedule" name="title"/>
                <FRBRdate date="1977" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRcountry value="na"/>
                <FRBRnumber value="25"/>
              </FRBRWork>
              <FRBRExpression>
                <FRBRthis value="/na/act/1977/25/eng@2021-01-01/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25/eng@2021-01-01"/>
                <FRBRdate date="2021-01-01" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRlanguage language="eng"/>
              </FRBRExpression>
              <FRBRManifestation>
                <FRBRthis value="/na/act/1977/25/eng@2021-01-01/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25/eng@2021-01-01"/>
                <FRBRdate date="2020-03-25" name="Generation"/>
                <FRBRauthor href=""/>
              </FRBRManifestation>
            </identification>
          </meta>''', self.tostring(components[1].doc.meta))

        self.assertEqual('''<meta xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
            <identification source="#cobalt">
              <FRBRWork>
                <FRBRthis value="/na/act/1977/25/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25"/>
                <FRBRalias value="Schedule" name="title"/>
                <FRBRdate date="1977" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRcountry value="na"/>
                <FRBRnumber value="25"/>
              </FRBRWork>
              <FRBRExpression>
                <FRBRthis value="/na/act/1977/25/eng@2021-01-01/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25/eng@2021-01-01"/>
                <FRBRdate date="2021-01-01" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRlanguage language="eng"/>
              </FRBRExpression>
              <FRBRManifestation>
                <FRBRthis value="/na/act/1977/25/eng@2021-01-01/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25/eng@2021-01-01"/>
                <FRBRdate date="2020-03-25" name="Generation"/>
                <FRBRauthor href=""/>
              </FRBRManifestation>
            </identification>
          </meta>''', self.tostring(components[2].doc.meta))

    def test_component_manifestation_date(self):
        self.a.manifestation_date = '2021-01-01'

        components = list(self.a.components().values())

        self.assertEqual('''<meta xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
            <identification source="#cobalt">
              <FRBRWork>
                <FRBRthis value="/na/act/1977/25/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25"/>
                <FRBRalias value="Schedule" name="title"/>
                <FRBRdate date="1977" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRcountry value="na"/>
                <FRBRnumber value="25"/>
              </FRBRWork>
              <FRBRExpression>
                <FRBRthis value="/na/act/1977/25/eng@1993-12-02/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                <FRBRdate date="1993-12-02" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRlanguage language="eng"/>
              </FRBRExpression>
              <FRBRManifestation>
                <FRBRthis value="/na/act/1977/25/eng@1993-12-02/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                <FRBRdate date="2021-01-01" name="Generation"/>
                <FRBRauthor href=""/>
              </FRBRManifestation>
            </identification>
          </meta>''', self.tostring(components[1].doc.meta))

        self.assertEqual('''<meta xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
            <identification source="#cobalt">
              <FRBRWork>
                <FRBRthis value="/na/act/1977/25/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25"/>
                <FRBRalias value="Schedule" name="title"/>
                <FRBRdate date="1980-01-01" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRcountry value="na"/>
                <FRBRnumber value="25"/>
              </FRBRWork>
              <FRBRExpression>
                <FRBRthis value="/na/act/1977/25/eng@1993-12-02/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                <FRBRdate date="1980-01-01" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRlanguage language="eng"/>
              </FRBRExpression>
              <FRBRManifestation>
                <FRBRthis value="/na/act/1977/25/eng@1993-12-02/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25/eng@1993-12-02"/>
                <FRBRdate date="2021-01-01" name="Generation"/>
                <FRBRauthor href=""/>
              </FRBRManifestation>
            </identification>
          </meta>''', self.tostring(components[2].doc.meta))

    def test_component_language(self):
        self.a.language = 'swa'

        components = list(self.a.components().values())

        self.assertEqual('''<meta xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
            <identification source="#cobalt">
              <FRBRWork>
                <FRBRthis value="/na/act/1977/25/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25"/>
                <FRBRalias value="Schedule" name="title"/>
                <FRBRdate date="1977" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRcountry value="na"/>
                <FRBRnumber value="25"/>
              </FRBRWork>
              <FRBRExpression>
                <FRBRthis value="/na/act/1977/25/swa@1993-12-02/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25/swa@1993-12-02"/>
                <FRBRdate date="1993-12-02" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRlanguage language="swa"/>
              </FRBRExpression>
              <FRBRManifestation>
                <FRBRthis value="/na/act/1977/25/swa@1993-12-02/!schedule-A"/>
                <FRBRuri value="/na/act/1977/25/swa@1993-12-02"/>
                <FRBRdate date="2020-03-25" name="Generation"/>
                <FRBRauthor href=""/>
              </FRBRManifestation>
            </identification>
          </meta>''', self.tostring(components[1].doc.meta))

        self.assertEqual('''<meta xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
            <identification source="#cobalt">
              <FRBRWork>
                <FRBRthis value="/na/act/1977/25/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25"/>
                <FRBRalias value="Schedule" name="title"/>
                <FRBRdate date="1977" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRcountry value="na"/>
                <FRBRnumber value="25"/>
              </FRBRWork>
              <FRBRExpression>
                <FRBRthis value="/na/act/1977/25/swa@1993-12-02/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25/swa@1993-12-02"/>
                <FRBRdate date="1980-01-01" name="Generation"/>
                <FRBRauthor href=""/>
                <FRBRlanguage language="swa"/>
              </FRBRExpression>
              <FRBRManifestation>
                <FRBRthis value="/na/act/1977/25/swa@1993-12-02/!schedule-XXX"/>
                <FRBRuri value="/na/act/1977/25/swa@1993-12-02"/>
                <FRBRdate date="2020-03-25" name="Generation"/>
                <FRBRauthor href=""/>
              </FRBRManifestation>
            </identification>
          </meta>''', self.tostring(components[2].doc.meta))



from unittest import TestCase

from cobalt import Judgment


class JudgmentTestCase(TestCase):
    maxDiff = None

    def test_empty_judgment(self):
        j = Judgment()
        self.assertEqual(j.title, "Untitled")
        self.assertIsNotNone(j.meta)
        self.assertIsNotNone(j.judgmentBody)

    def test_empty_body(self):
        j = Judgment()
        self.assertNotEqual(j.judgmentBody.text, '')

    def test_main(self):
        j = Judgment()
        self.assertEqual(j.main, j.judgment)



from unittest import TestCase
from datetime import date

from lxml.etree import LxmlSyntaxError

from cobalt import Act, AmendmentEvent, RepealEvent, datestring
from cobalt.schemas import assert_validates


class ActTestCase(TestCase):
    maxDiff = None

    def test_empty_act(self):
        a = Act()
        self.assertEqual(a.title, "Untitled")
        self.assertIsNotNone(a.meta)
        self.assertIsNotNone(a.body)

        assert_validates(a)

    def test_empty_body(self):
        a = Act()
        self.assertNotEqual(a.body.text, '')
        assert_validates(a)

    def test_publication_date(self):
        a = Act()
        self.assertIsNone(a.publication_date)
        assert_validates(a)

        a.publication_date = '2012-01-02'
        self.assertEqual(datestring(a.publication_date), '2012-01-02')
        self.assertIsInstance(a.publication_date, date)
        assert_validates(a)

        a.publication_date = None
        self.assertIsNone(a.publication_date)
        assert_validates(a)

    def test_publication_number(self):
        a = Act()
        self.assertIsNone(a.publication_number)

        a.publication_number = '1234'
        self.assertEqual(a.publication_number, '1234')
        assert_validates(a)

    def test_publication_name(self):
        a = Act()
        self.assertIsNone(a.publication_name)

        a.publication_name = 'Publication'
        self.assertEqual(a.publication_name, 'Publication')
        assert_validates(a)

    def test_set_amendments(self):
        a = Act()
        a.frbr_uri = "/akn/za/act/1900/1"
        a.amendments = [AmendmentEvent(date='2012-02-01', amending_uri='/za/act/1980/10', amending_title="Foo")]
        assert_validates(a)

        xml = a.to_xml(encoding='unicode', pretty_print=True)
        xml = xml.replace(datestring(date.today()), 'TODAY')

        self.assertMultiLineEqual(
            """<akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0">
  <act name="act" contains="singleVersion">
    <meta>
      <identification source="#cobalt">
        <FRBRWork>
          <FRBRthis value="/akn/za/act/1900/1/!main"/>
          <FRBRuri value="/akn/za/act/1900/1"/>
          <FRBRalias value="Untitled" name="title"/>
          <FRBRdate date="1900" name="Generation"/>
          <FRBRauthor href=""/>
          <FRBRcountry value="za"/>
          <FRBRnumber value="1"/>
        </FRBRWork>
        <FRBRExpression>
          <FRBRthis value="/akn/za/act/1900/1/eng@TODAY/!main"/>
          <FRBRuri value="/akn/za/act/1900/1/eng@TODAY"/>
          <FRBRdate date="TODAY" name="Generation"/>
          <FRBRauthor href=""/>
          <FRBRlanguage language="eng"/>
        </FRBRExpression>
        <FRBRManifestation>
          <FRBRthis value="/akn/za/act/1900/1/eng@TODAY/!main"/>
          <FRBRuri value="/akn/za/act/1900/1/eng@TODAY"/>
          <FRBRdate date="TODAY" name="Generation"/>
          <FRBRauthor href=""/>
        </FRBRManifestation>
      </identification>
      <lifecycle source="#cobalt">
        <eventRef eId="amendment-2012-02-01" date="2012-02-01" type="amendment" source="#amendment-0-source"/>
      </lifecycle>
      <references source="#cobalt">
        <TLCOrganization eId="cobalt" href="https://github.com/laws-africa/cobalt" showAs="cobalt"/>
        <passiveRef eId="amendment-0-source" href="/za/act/1980/10" showAs="Foo"/>
      </references>
    </meta>
    <body>
      <section eId="sec_nn_1">
        <content>
          <p eId="sec_nn_1__p_1"/>
        </content>
      </section>
    </body>
  </act>
</akomaNtoso>
""", xml)

        a.amendments = [
            AmendmentEvent(date='2012-02-01', amending_uri='/za/act/1980/22', amending_title="Corrected"),
            AmendmentEvent(date='2013-03-03', amending_uri='/za/act/1990/5', amending_title="Bar"),
        ]
        assert_validates(a)

        xml = a.to_xml(encoding='unicode', pretty_print=True)
        xml = xml.replace(datestring(date.today()), 'TODAY')

        self.assertMultiLineEqual(
            """<akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0">
  <act name="act" contains="singleVersion">
    <meta>
      <identification source="#cobalt">
        <FRBRWork>
          <FRBRthis value="/akn/za/act/1900/1/!main"/>
          <FRBRuri value="/akn/za/act/1900/1"/>
          <FRBRalias value="Untitled" name="title"/>
          <FRBRdate date="1900" name="Generation"/>
          <FRBRauthor href=""/>
          <FRBRcountry value="za"/>
          <FRBRnumber value="1"/>
        </FRBRWork>
        <FRBRExpression>
          <FRBRthis value="/akn/za/act/1900/1/eng@TODAY/!main"/>
          <FRBRuri value="/akn/za/act/1900/1/eng@TODAY"/>
          <FRBRdate date="TODAY" name="Generation"/>
          <FRBRauthor href=""/>
          <FRBRlanguage language="eng"/>
        </FRBRExpression>
        <FRBRManifestation>
          <FRBRthis value="/akn/za/act/1900/1/eng@TODAY/!main"/>
          <FRBRuri value="/akn/za/act/1900/1/eng@TODAY"/>
          <FRBRdate date="TODAY" name="Generation"/>
          <FRBRauthor href=""/>
        </FRBRManifestation>
      </identification>
      <lifecycle source="#cobalt">
        <eventRef eId="amendment-2012-02-01" date="2012-02-01" type="amendment" source="#amendment-0-source"/>
        <eventRef eId="amendment-2013-03-03" date="2013-03-03" type="amendment" source="#amendment-1-source"/>
      </lifecycle>
      <references source="#cobalt">
        <TLCOrganization eId="cobalt" href="https://github.com/laws-africa/cobalt" showAs="cobalt"/>
        <passiveRef eId="amendment-0-source" href="/za/act/1980/22" showAs="Corrected"/>
        <passiveRef eId="amendment-1-source" href="/za/act/1990/5" showAs="Bar"/>
      </references>
    </meta>
    <body>
      <section eId="sec_nn_1">
        <content>
          <p eId="sec_nn_1__p_1"/>
        </content>
      </section>
    </body>
  </act>
</akomaNtoso>
""", xml)

        amendment = a.amendments[0]
        self.assertEqual(datestring(amendment.date), '2012-02-01')
        self.assertEqual(amendment.amending_uri, '/za/act/1980/22')
        self.assertEqual(amendment.amending_title, 'Corrected')

        amendment = a.amendments[1]
        self.assertEqual(datestring(amendment.date), '2013-03-03')
        self.assertEqual(amendment.amending_uri, '/za/act/1990/5')
        self.assertEqual(amendment.amending_title, 'Bar')

        assert_validates(a)

        # clear them
        a.amendments = []
        assert_validates(a)
        xml = a.to_xml(encoding='unicode', pretty_print=True)
        xml = xml.replace(datestring(date.today()), 'TODAY')

        self.assertMultiLineEqual(
            """<akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0">
  <act name="act">
    <meta>
      <identification source="#cobalt">
        <FRBRWork>
          <FRBRthis value="/akn/za/act/1900/1/!main"/>
          <FRBRuri value="/akn/za/act/1900/1"/>
          <FRBRalias value="Untitled" name="title"/>
          <FRBRdate date="1900" name="Generation"/>
          <FRBRauthor href=""/>
          <FRBRcountry value="za"/>
          <FRBRnumber value="1"/>
        </FRBRWork>
        <FRBRExpression>
          <FRBRthis value="/akn/za/act/1900/1/eng@TODAY/!main"/>
          <FRBRuri value="/akn/za/act/1900/1/eng@TODAY"/>
          <FRBRdate date="TODAY" name="Generation"/>
          <FRBRauthor href=""/>
          <FRBRlanguage language="eng"/>
        </FRBRExpression>
        <FRBRManifestation>
          <FRBRthis value="/akn/za/act/1900/1/eng@TODAY/!main"/>
          <FRBRuri value="/akn/za/act/1900/1/eng@TODAY"/>
          <FRBRdate date="TODAY" name="Generation"/>
          <FRBRauthor href=""/>
        </FRBRManifestation>
      </identification>
      <references source="#cobalt">
        <TLCOrganization eId="cobalt" href="https://github.com/laws-africa/cobalt" showAs="cobalt"/>
      </references>
    </meta>
    <body>
      <section eId="sec_nn_1">
        <content>
          <p eId="sec_nn_1__p_1"/>
        </content>
      </section>
    </body>
  </act>
</akomaNtoso>
""", xml)

    def test_set_repeal(self):
        a = Act()
        a.frbr_uri = "/akn/za/act/1900/1"
        a.repeal = RepealEvent(date='2012-02-01', repealing_uri='/za/act/1980/10', repealing_title='Foo')
        xml = a.to_xml(encoding='unicode', pretty_print=True)
        xml = xml.replace(datestring(date.today()), 'TODAY')

        self.assertMultiLineEqual(
            """<akomaNtoso xmlns="http://docs.oasis-open.org/legaldocml/ns/akn/3.0">
  <act name="act">
    <meta>
      <identification source="#cobalt">
        <FRBRWork>
          <FRBRthis value="/akn/za/act/1900/1/!main"/>
          <FRBRuri value="/akn/za/act/1900/1"/>
          <FRBRalias value="Untitled" name="title"/>
          <FRBRdate date="1900" name="Generation"/>
          <FRBRauthor href=""/>
          <FRBRcountry value="za"/>
          <FRBRnumber value="1"/>
        </FRBRWork>
        <FRBRExpression>
          <FRBRthis value="/akn/za/act/1900/1/eng@TODAY/!main"/>
          <FRBRuri value="/akn/za/act/1900/1/eng@TODAY"/>
          <FRBRdate date="TODAY" name="Generation"/>
          <FRBRauthor href=""/>
          <FRBRlanguage language="eng"/>
        </FRBRExpression>
        <FRBRManifestation>
          <FRBRthis value="/akn/za/act/1900/1/eng@TODAY/!main"/>
          <FRBRuri value="/akn/za/act/1900/1/eng@TODAY"/>
          <FRBRdate date="TODAY" name="Generation"/>
          <FRBRauthor href=""/>
        </FRBRManifestation>
      </identification>
      <lifecycle source="#cobalt">
        <eventRef eId="repeal-2012-02-01" date="2012-02-01" type="repeal" source="#repeal-source"/>
      </lifecycle>
      <references source="#cobalt">
        <TLCOrganization eId="cobalt" href="https://github.com/laws-africa/cobalt" showAs="cobalt"/>
        <passiveRef eId="repeal-source" href="/za/act/1980/10" showAs="Foo"/>
      </references>
    </meta>
    <body>
      <section eId="sec_nn_1">
        <content>
          <p eId="sec_nn_1__p_1"/>
        </content>
      </section>
    </body>
  </act>
</akomaNtoso>
""", xml)

        self.assertEqual(a.repeal.repealing_uri, '/za/act/1980/10')
        self.assertEqual(a.repeal.repealing_title, 'Foo')
        self.assertEqual(datestring(a.repeal.date), '2012-02-01')

        assert_validates(a)

        # check that clearing it works
        a.repeal = None
        self.assertIsNone(a.repeal)

        assert_validates(a)

    def test_main(self):
        a = Act()
        self.assertEqual(a.main, a.act)

    def test_bad_xml(self):
        with self.assertRaises(LxmlSyntaxError):
            Act('badness')

        with self.assertRaises(ValueError):
            Act('<root>no namespace</root>')

        with self.assertRaises(ValueError):
            Act('<akomaNtoso>no first child</akomaNtoso>')

        with self.assertRaises(ValueError):
            Act('<akomaNtoso><child>no namespace</child></akomaNtoso>')



from unittest import TestCase

from cobalt.uri import FrbrUri


class FrbrUriTestCase(TestCase):
    def test_bad_value(self):
        self.assertRaises(ValueError, FrbrUri.parse, "/badness")
        self.assertRaises(ValueError, FrbrUri.parse, "/ukpga/2015/1")

    def test_simple(self):
        uri = FrbrUri.parse("/akn/za/act/1980/01")
        self.assertEqual(uri.country, "za")
        self.assertEqual(uri.locality, None)
        self.assertEqual(uri.doctype, "act")
        self.assertEqual(uri.subtype, None)
        self.assertEqual(uri.actor, None)
        self.assertEqual(uri.date, "1980")
        self.assertEqual(uri.number, "01")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, None)
        self.assertEqual(uri.prefix, "akn")

        self.assertEqual("/akn/za/act/1980/01", uri.work_uri())

    def test_with_subtype(self):
        uri = FrbrUri.parse("/akn/za/act/by-law/1980/01")
        self.assertEqual(uri.country, "za")
        self.assertEqual(uri.locality, None)
        self.assertEqual(uri.doctype, "act")
        self.assertEqual(uri.subtype, "by-law")
        self.assertEqual(uri.actor, None)
        self.assertEqual(uri.date, "1980")
        self.assertEqual(uri.number, "01")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, None)

        self.assertEqual("/akn/za/act/by-law/1980/01", uri.work_uri())

    def test_with_locality(self):
        uri = FrbrUri.parse("/akn/za-cpt/act/by-law/1980/01")
        self.assertEqual(uri.country, "za")
        self.assertEqual(uri.locality, "cpt")
        self.assertEqual(uri.doctype, "act")
        self.assertEqual(uri.subtype, "by-law")
        self.assertEqual(uri.actor, None)
        self.assertEqual(uri.date, "1980")
        self.assertEqual(uri.number, "01")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, None)

        self.assertEqual("/akn/za-cpt/act/by-law/1980/01", uri.work_uri())

    def test_with_subtype_and_actor(self):
        uri = FrbrUri.parse("/akn/za/act/by-law/actor/1980/01")
        self.assertEqual(uri.country, "za")
        self.assertEqual(uri.doctype, "act")
        self.assertEqual(uri.subtype, "by-law")
        self.assertEqual(uri.actor, "actor")
        self.assertEqual(uri.date, "1980")
        self.assertEqual(uri.number, "01")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, None)

        self.assertEqual("/akn/za/act/by-law/actor/1980/01", uri.work_uri())

    def test_with_long_date(self):
        uri = FrbrUri.parse("/akn/za/act/1980-02-01/01")
        self.assertEqual(uri.country, "za")
        self.assertEqual(uri.doctype, "act")
        self.assertEqual(uri.subtype, None)
        self.assertEqual(uri.actor, None)
        self.assertEqual(uri.date, "1980-02-01")
        self.assertEqual(uri.year, "1980")
        self.assertEqual(uri.number, "01")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, None)

        self.assertEqual("/akn/za/act/1980-02-01/01", uri.work_uri())

    def test_with_non_numeric_number(self):
        uri = FrbrUri.parse("/akn/za/act/1980/nn")
        self.assertEqual(uri.country, "za")
        self.assertEqual(uri.doctype, "act")
        self.assertEqual(uri.subtype, None)
        self.assertEqual(uri.actor, None)
        self.assertEqual(uri.date, "1980")
        self.assertEqual(uri.number, "nn")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, None)

        self.assertEqual("/akn/za/act/1980/nn", uri.work_uri())

    def test_with_work_component(self):
        uri = FrbrUri.parse("/akn/za/act/1980/2/!schedule1")
        self.assertEqual(uri.country, "za")
        self.assertEqual(uri.doctype, "act")
        self.assertEqual(uri.subtype, None)
        self.assertEqual(uri.actor, None)
        self.assertEqual(uri.date, "1980")
        self.assertEqual(uri.number, "2")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.work_component, "schedule1")
        self.assertEqual(uri.expression_date, None)

        self.assertEqual("/akn/za/act/1980/2", uri.uri())
        self.assertEqual("/akn/za/act/1980/2/!schedule1", uri.work_uri())
        self.assertEqual("/akn/za/act/1980/2/eng/!schedule1", uri.expression_uri())

    def test_with_nested_work_components(self):
        uri = FrbrUri.parse("/akn/za/act/1980/2/!schedule1/schedule2/schedule3")
        self.assertEqual(uri.work_component, "schedule1/schedule2/schedule3")

    def test_with_work_component_legacy(self):
        uri = FrbrUri.parse("/akn/za/act/1980/2/!schedule1")
        self.assertEqual(uri.country, "za")
        self.assertEqual(uri.doctype, "act")
        self.assertEqual(uri.subtype, None)
        self.assertEqual(uri.actor, None)
        self.assertEqual(uri.date, "1980")
        self.assertEqual(uri.number, "2")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.work_component, "schedule1")
        self.assertEqual(uri.expression_date, None)

        self.assertEqual("/akn/za/act/1980/2", uri.uri())
        self.assertEqual("/akn/za/act/1980/2/!schedule1", uri.work_uri())
        self.assertEqual("/akn/za/act/1980/2/eng/!schedule1", uri.expression_uri())

    def test_with_short_work_component(self):
        uri = FrbrUri.parse("/akn/za-wc/act/pn/2018/46/!6")
        self.assertEqual(uri.country, "za")
        self.assertEqual(uri.locality, "wc")
        self.assertEqual(uri.doctype, "act")
        self.assertEqual(uri.subtype, "pn")
        self.assertEqual(uri.actor, None)
        self.assertEqual(uri.date, "2018")
        self.assertEqual(uri.number, "46")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.work_component, "6")
        self.assertEqual(uri.expression_date, None)

        uri = FrbrUri.parse("/akn/za-wc/act/2018/46/!6")
        self.assertEqual(uri.country, "za")
        self.assertEqual(uri.locality, "wc")
        self.assertEqual(uri.doctype, "act")
        self.assertEqual(uri.subtype, None)
        self.assertEqual(uri.actor, None)
        self.assertEqual(uri.date, "2018")
        self.assertEqual(uri.number, "46")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.work_component, "6")
        self.assertEqual(uri.expression_date, None)

    def test_with_work_component_and_portion(self):
        uri = FrbrUri.parse("/akn/za/act/1980/2")
        uri.work_component = "main/schedule_1"
        uri.portion = "chp_2"

        self.assertEqual("/akn/za/act/1980/2", uri.uri())
        self.assertEqual("/akn/za/act/1980/2/!main/schedule_1", uri.work_uri())
        self.assertEqual("/akn/za/act/1980/2/eng/!main/schedule_1~chp_2", uri.expression_uri())

    def test_parse_expression2(self):
        uri = FrbrUri.parse("/gh/act/2020/1013/eng@2020-04-03")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, '@2020-04-03')
        self.assertEqual(uri.number, '1013')
        self.assertIsNone(uri.subtype)

    def test_expression_string_no_language(self):
        uri = FrbrUri.parse("/gh/act/2020/1013/")
        self.assertEqual(uri.language, "eng")
        uri.language = None
        with self.assertRaises(ValueError) as e:
            uri.expression_uri()
        err = e.exception
        self.assertEqual(str(err), "Expression URI requires a language.")

    def test_parse_expression(self):
        uri = FrbrUri.parse("/akn/za/act/1980/02/afr@")
        self.assertEqual(uri.language, "afr")
        self.assertEqual(uri.expression_date, '@')
        self.assertEqual("/akn/za/act/1980/02", uri.work_uri())
        self.assertEqual("/akn/za/act/1980/02/afr@", uri.expression_uri())

        uri = FrbrUri.parse("/akn/za/act/1980/02/afr@2014-01-01")
        self.assertEqual(uri.language, "afr")
        self.assertEqual(uri.expression_date, "@2014-01-01")
        self.assertEqual("/akn/za/act/1980/02", uri.work_uri())
        self.assertEqual("/akn/za/act/1980/02/afr@2014-01-01", uri.expression_uri())

        uri = FrbrUri.parse("/akn/za/act/1980/02/afr.html")
        self.assertEqual(uri.language, "afr")
        self.assertEqual(uri.format, 'html')

    def test_parse_expression_component_legacy(self):
        """ Legacy components without a ! are no longer supported.
        """
        with self.assertRaises(ValueError):
            FrbrUri.parse("/akn/za/act/1980/02/eng/main")

    def test_parse_work_component(self):
        uri = FrbrUri.parse("/akn/za/act/1980/02/eng/!main")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, None)
        self.assertEqual(uri.work_component, "main")

        uri = FrbrUri.parse("/akn/za/act/1980/02/eng/!main~chp_2")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, None)
        self.assertEqual(uri.work_component, "main")
        self.assertEqual(uri.portion, "chp_2")

        uri = FrbrUri.parse("/akn/za/act/1980/02/eng@/!main~chp_2")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, '@')
        self.assertEqual(uri.work_component, "main")
        self.assertEqual(uri.portion, "chp_2")

    def test_parse_expression_date(self):
        # A dangling @ indicates the very FIRST expression date, which
        # we represent with an empty string ('').
        # A URI without an @ at all, indicates the most recent
        # expression date, which is None.

        uri = FrbrUri.parse("/akn/za/act/1980/02/eng")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, None)
        self.assertEqual(uri.expression_uri(), '/akn/za/act/1980/02/eng')

        uri = FrbrUri.parse("/akn/za/act/1980/02/eng/!main")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, None)

        uri = FrbrUri.parse("/akn/za/act/1980/02/eng@")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, '@')
        self.assertEqual(uri.expression_uri(), '/akn/za/act/1980/02/eng@')

        uri = FrbrUri.parse("/akn/za/act/1980/02/eng@/!main")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, '@')

        uri = FrbrUri.parse("/akn/za/act/1980/02/eng:/!main")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, ':')

        uri = FrbrUri.parse("/akn/za/act/1980/02/eng:2012-01-01/!main")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, ':2012-01-01')

    def test_parse_subtype_numeric_number(self):
        # A subtype with a numeric number should not be
        # mistaken for an actor
        uri = FrbrUri.parse("/akn/za-jhb/act/notice/2007/5319/eng@2007-12-05")
        self.assertIsNone(uri.actor)
        self.assertEqual(uri.date, "2007")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.number, "5319")
        self.assertEqual(uri.expression_date, "@2007-12-05")

    def test_parse_subtype_and_actor(self):
        uri = FrbrUri.parse("/akn/za-jhb/act/notice/actor/2007/5319/eng@2007-12-05")
        self.assertEqual(uri.actor, "actor")
        self.assertEqual(uri.date, "2007")
        self.assertEqual(uri.number, "5319")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, "@2007-12-05")

    def test_expression_uri(self):
        uri = FrbrUri.parse("/akn/za/act/1980/02/eng")
        uri.expression_date = '@2014-01-01'
        uri.work_component = 'main'
        uri.format = 'html'

        self.assertEqual("/akn/za/act/1980/02/eng@2014-01-01/!main", uri.expression_uri())

    def test_manifestation_uri(self):
        uri = FrbrUri.parse("/akn/za/act/1980/02/eng")
        uri.expression_date = '@2014-01-01'
        uri.work_component = 'main'
        uri.format = 'html'

        self.assertEqual("/akn/za/act/1980/02/eng@2014-01-01/!main.html", uri.manifestation_uri())

    def test_simple_prefix(self):
        # also recognises akn prefix
        uri = FrbrUri.parse("/akn/za/act/1980/01")
        self.assertEqual(uri.prefix, "akn")
        self.assertEqual(uri.country, "za")
        self.assertEqual(uri.locality, None)
        self.assertEqual(uri.doctype, "act")
        self.assertEqual(uri.subtype, None)
        self.assertEqual(uri.actor, None)
        self.assertEqual(uri.date, "1980")
        self.assertEqual(uri.number, "01")
        self.assertEqual(uri.language, "eng")
        self.assertEqual(uri.expression_date, None)

        self.assertEqual("/akn/za/act/1980/01", uri.work_uri())

    def test_akn_prefix(self):
        # uses 'akn' prefix as the default when producing one
        uri = FrbrUri(
            country='za',
            locality='ec',
            doctype='act',
            subtype='by-law',
            date='2020',
            number='31',
            actor=None
        )
        self.assertEqual(uri.prefix, "akn")

        # doesn't use 'akn' prefix if explicitly told not to
        uri = FrbrUri(
            prefix=None,
            country='za',
            locality='ec',
            doctype='act',
            subtype='by-law',
            date='2020',
            number='31',
            actor=None
        )
        self.assertIsNone(uri.prefix)

    def test_parse_portion_no_component(self):
        uri = FrbrUri.parse("/akn/za/act/2005/5/~sec_5")
        self.assertEqual(uri.portion, "sec_5")

        uri = FrbrUri.parse("/akn/za/act/2005/5/eng/~sec_5")
        self.assertEqual(uri.portion, "sec_5")

        uri = FrbrUri.parse("/akn/za/act/2005/5/eng@2002-03-01/~sec_5")
        self.assertEqual(uri.portion, "sec_5")
        self.assertEqual(uri.expression_uri(), "/akn/za/act/2005/5/eng@2002-03-01/~sec_5")

    def test_parse_portion_no_component_format(self):
        uri = FrbrUri.parse("/akn/za/act/2005/5/~sec_5.html")
        self.assertEqual(uri.portion, "sec_5")
        self.assertEqual(uri.format, "html")

        uri = FrbrUri.parse("/akn/za/act/2005/5/eng/~sec_5.xml")
        self.assertEqual(uri.portion, "sec_5")
        self.assertEqual(uri.format, "xml")

        uri = FrbrUri.parse("/akn/za/act/2005/5/eng@2002-03-01/~sec_5.xml")
        self.assertEqual(uri.portion, "sec_5")
        self.assertEqual(uri.format, "xml")

    def test_parse_portion_component(self):
        uri = FrbrUri.parse("/akn/za/act/2005/5/!main~sec_5")
        self.assertEqual(uri.work_component, "main")
        self.assertEqual(uri.portion, "sec_5")

        uri = FrbrUri.parse("/akn/za/act/2005/5/eng/!schedule_3~sec_5")
        self.assertEqual(uri.work_component, "schedule_3")
        self.assertEqual(uri.portion, "sec_5")

        uri = FrbrUri.parse("/akn/za/act/2005/5/eng@2002-03-01/!main~sec_5")
        self.assertEqual(uri.work_component, "main")
        self.assertEqual(uri.portion, "sec_5")
